#! perl -w

=head1 NAME

pythia - GloMop Project HTTP-compliant distillation proxy

=head1 SYNOPSIS

Pythia is a proxy that performs distillation of images, HTML, and other
data for commodity browsers running on impoverished devices or over an
impoverished connection.  It is one of the software components developed
as part of the UC Berkeley GloMop research project
(http://www.cs.berkeley.edu/~fox/glomop).  Pythia can easily be extended to
experiment with other types of computing in the infrastructure.

Pythia requires the MD5 and LWP Perl modules, both of which are freely
available from your local CPAN mirror.  If you don't know what this is,
try C<http://www.metronet.com/perlinfo/> for help.

To start Pythia, say "C<pythia.pl -help>" to see runtime options.    

This code is really intended as an experimental base for people to hack on and
build their own services, etc. on top of.  To make that easier, each of the
various Perl files that makes up Pythia has its own man page.  Read the
overview below to understand the API's exported by Pythia, and what is
contained in the various other files.

=head1 DESCRIPTION

Pythia exports two API's of interest, each with its own man page:

=over 3

=item *

HTML distillation and page "segmenting": L<munge_html>.  Using this API also
requires some familiarity with L<html_parser>.  You need this API if you
want to be able to transform HTML on the fly, e.g. to change the
destinations of IMG tags, remove or substitute tags, etc.

=item *

Distillation service API for other MIME types: L<services>.

=back

In addition, Pythia is made up of several hackable components, each
of which is described in its own man page; read the man pages, Use The Source,
and program by example, and you should have a fine time.  If you want to
contribute more detailed documentation, I'd be happy to have it.  :-)

=over 3

=item L<child>

Per-connection code for Pythia.  The "lifetime of a single HTTP request"
control path can be found here.

=item L<html_parser>

Wrapper around the excellent HTML parser found in the LWP (libwww-perl)
package.  Read this if you intend to muck around with HTML on-the-fly
modification, which Pythia performs to allow refinement and other cool
things. 

=item L<advice>

Receiving external notifications that interesting properties
of the client device or its network have changed, thus allowing
automatic adaptation to those changes.

=item L<cache>

A brain-dead local cache to assist distillation, since Pythia was
developed in an environment where no Web caches were deployed.

=item L<client>

A way to track client-device-specific properties and use them to make
distillation decisions.

=item L<prefs>

Simple user preferences database management.

=item L<register.html>

Support for sending the user an HTML form that can be used to change
preferences. 

=item L<errors>

Reporting errors to the user via HTTP/HTML.

=item L<gm_sock>

Usable abstractions for sockets.

=item L<gif> and L<jpeg>

Two distillers that come with Pythia.  These files describe the Perl
code that makes the "distillation decision"; the distillation itself is
executed by C programs L<gifmunch> and cjpeg/djpeg, subject to the
parameters chosen by the decision function.

=item 

Play around with it and let me know of any interesting results.

=head1 REFERENCES

Armando Fox and Eric A. Brewer, I<Reducing WWW Latency and Bandwidth
Requirements via Real-Time Distillation>.  Proc. Fifth International
World Wide Web Conference (WWW-5), Paris, May 1996.

Armando Fox, Steven D. Gribble, Eric A. Brewer, and Elan Amir.
I<Adapting to Network and Client Variability via On-Demand Dynamic
Transcoding>.  Proc. Seventh International Conference on Architectural
Support for Programming Languages and Operating Systems (ASPLOS-VII),
Boston, October 1996 (to appear).

=head1 BUGS

Probably numerous ones, but as you know from reading the License, this
code is offered as is, etc.

Architectural bug: HTML distillation and other distillation should all
use the same API.

Architectural bug:  the distinction between user preferences and client
device profile is nebulous.

=head1 COPYRIGHT

All the Pythia code is Copyright (C) 1994-1996, the Regents of the University
of California.  Read the LICENSE file for license terms. 

=cut


use POSIX;
use MD5;
use LWP;


#
#  Per-connection "globals"
#
sub instvars {
    my $class = shift;
    foreach $var (@_) {
        if ($var =~ /(\S+):r/) {
            # read-only variable
            $var = $1;
            $action = qq!carp "Can't set instance variable `$var' this way, ign\
ored"!;
        } else {
            $action = qq!\$self->{"$var"} = shift!;
        }
        eval "sub ${class}::${var}"  . qq!{
            my \$self = shift;
            $action if defined \$_[0];
            return \$self->{"$var"}; }! ;
    }
}

&instvars('conn', qw(IPaddr URL method proto request reqtype sock data resp));

sub conn::new {
    bless {
        "IPaddr"        => undef,
        "URL"           => undef,
        "proto"         => undef,
        "request"       => undef,
        "reqtype"       => undef,
        "client_headers" => {},
        "sock"          => undef,
        "data"          => undef,
        "resp"          => undef,
    }
}

#
# Globals reused frequently
#

%Children = ();                             # outstanding unreaped children
$Listen = '';                               # listening socket object
$Accept = '';                               # accepted socket object

#
# defaults  for cmd line options
#
$gb{neverfork} = 0;
$help{neverfork} = 'Run in debugging mode: never fork';

$g{statfile} = '/dev/null';
$help{statfile} = 'File in which to write distillation statistics';

$g{port}  = 5555;                     # http port
$help{port} = 'HTTP port';

$g{sport} = 9100;
$help{sport} = 'Default port for starting local services';

$g{aport} = 9542;
$help{aport} = 'Default port for UDP event-notification packets';

$g{qsize} = 10;
$help{qsize} = 'Socket queue size';

$g{gateway} = '';
$help{gateway} = 'Firewall gateway machine (host:port) (CURRENTLY BROKEN)';

$g{config} = './pythia.cf';
$help{config} = 'Pythia configuration file';

$gb{keeptemp} = 0;
$help{keeptemp} = "Don't delete temp files after exiting";

#
#  startup
#

foreach $file (qw(gm_sock
                  html_parser
                  child
                  errors
                  prefs
                  services
                  client
                  register.html
                  cache
                  munge_html
                  advice
                  ismap_icon
                  imagemap
                  ))  {
    require "${file}.pl";
}

&parse_args;

#
#  Open a socket for listening.
#
($Listen = listensock gm_sock $g{port}, 'tcp', $g{qsize})
    or die("listensock (TCP): " . &gm_sock::lasterror);

#
#  And a UDP socket for advice.
#
($Advice = listensock gm_sock $g{aport}, 'udp')
    or die ("listensock (UDP): " . &gm_sock::lasterror);

&init;
&mainloop;

#------------------------------------------------------------
#  main loop: listen for connections, and fork child to handle them
#

sub mainloop {
    my($addr);
    my($c,$zz,$dummy);
    my($rbits,$ebits) = ('','');
    my($rout,$eout) = ();
    my($vec);
    my($from, $hmsg);
    my($url, $method, $proto, $reqtype, @hdrs);

    my($sockfh) = ($Listen->filehandles)[0];
    vec($rbits, fileno($sockfh), 1) = 1;
    vec($ebits, fileno($sockfh), 1) = 1;

    my($advicefh) = ($Advice->filehandles)[0];
    vec($rbits, fileno($advicefh), 1) = 1;
    vec($ebits, fileno($advicefh), 1) = 1;
    
    &log(sprintf("Listening on port %d/tcp, advice on %d/udp",
                 ($Listen->endpoints)[3], ($Advice->endpoints)[3]));

    CON: for ($Con = 0; ; $Con++) {

        $Accept->closesock() if (ref($Accept) eq gm_sock);

      SELECT: while (1) {
          #
          # Idle loop: while there are outstanding children, loop between
          # waiting on select() and reaping the children.  NOTE: Real POSIX
          # waitpid should allow us to specify 0 as the first arg, to check
          # for status of other processes in our process group, but it
          # doesn't seem to work in perl....
          #
          unless ($gb{neverfork}) {
              $zz = 1;
              while ($zz > 0) {
                  # BUG::for some reason the perl installation at geoworks
                  # cannot properly import waitpid from auto/POSIX/waitpid.al.
                  # Happily, under solaris, waitpid is provided directly, so we
                  # don't need to explicitly import it from the POSIX package.
                  $zz = waitpid(-1, &WNOHANG);
                  (delete $Children{$zz}, &debug('fork', "Reaped $zz"))
                      if $zz > 0;
              }
          }
          #$sel = select($rout=$rbits,undef,$eout=$ebits,$Const{SelectTimeout});
          $sel = select($rout=$rbits,undef,$eout=$ebits,undef);
          if ($sel == -1) { &log("*** select() error: $!"); }
          last SELECT if (($sel > 0) &&
                          ((vec($rout,fileno($sockfh),1)) |
                           (vec($rout,fileno($advicefh),1))));
      }

        # if this is a notification, handle it and then continue.
        
        if (vec($rout, fileno($advicefh), 1)) {
            if (($from,$hmsg) = $Advice->readsock_raw(65536)) {
                &debug("advice", "Advice received from $from");
                &advice::do_advice($hmsg);
            } else {
                &log("** Advice socket woke up, but no msg seen!");
            }
            next CON;
        }
          
        next CON unless (vec($rout, fileno($sockfh), 1));

        $Accept = $Listen->acceptsock();
        unless ($Accept) {
            &log( "*** accept: " . $Listen->lasterror());
            next CON;
        }
        # get the client's IP addr.
        $con = new conn;
        $con->{"sock"} = $Accept;
        $con->{"IPaddr"} = ($Accept->endpoints())[0];

        # collect HTTP request line.
        
        unless (&collect_request($con)) {
            &log("*** Bad request: " . $con->{"request"});
            &http_error($con,"500 Bad request");
            next CON;
        }

        #  Determine request type.  Possible requests:
        #   New service added or removed:
        #        SVCADD servicename@host:port
        #        SVCDEL servicename@host:port
        #   Registration
        #        GET url (where url is the magic "registration" URL)
        #   Prefs change
        #        GET url (where url is the "generate prefs page" URL)
        #   Stats query
        #        GET url (where url is "magic proxystats URL")
        #   Regular http

        #
        # If a new service is being added:
        #
        if ($con->{"method"} =~ /svcadd/i) {
            # the "URL" encodes the service name and location, e.g.
            #    "gifmunch@foobar.com:9999"
            my $svc = $con->{"URL"};
            my $ipad = ($Accept->endpoints())[0];
            my $port = $con->{"proto"};
            &distiller::birth($ipad,$port,$svc);
            next CON;
        }

        #
        # If a new service is being removed:
        #
        if ($con->{"method"} =~ /svcdel/i) {
            my $svc = $con->{"URL"};
            my $port = $con->{"proto"};
            my $ipad = ($Accept->endpoints())[0];
            &distiller::death($ipad, $port, $svc);
            next CON;
        }

        #
        # If this is a login:
        #
        if ($con->{"URL"} =~ s/^$main::Const{LoginURL}//o) {
            $con->{"sock"} = $Accept;
            &debug('register', "Login request from " . $con->{"IPaddr"});
            # login function will return HTML text giving status of login
            # transaction.
            &send_status($con, "200 OK (Pythia login)");
            &send_html($con,&login($con));
            next CON;
        }

        #
        # If this is a change of prefs:
        #
        if ($con->{"URL"} =~ s/^$main::Const{MatchRegisterURL}//o) {
            my $username = $1;
            my($x);
            $con->{"sock"} = $Accept;
            &debug('register',
                   join("Registration request from",
                        $con->{"IPaddr"}, "(user $username):",
                        $con->{"URL"}));
            # call register function, which returns HTML text giving status of
            # registration transaction.
            $x = &register($con->{"URL"}, $con->{"IPaddr"}, $username) ;
            if (defined $x) {
                &send_status($con, "200 OK (Pythia prefs change)");
                &send_html($con, $x);
            } else {
                # registration failed!
                &client_disallowed($con);
            }
            next CON;
        }

        #
        # If this is a stats request, return some stats.
        #

        if ($con->{"URL"} =~ /^$main::Const{StatsURL}$/o) {
            $Accept->writesock( &get_stats($con));
            next CON;
        }

        #
        # If this is a request for a Prefs Change form or Login form:
        #
        if ($con->{"URL"} =~ /$main::Const{RegisterRequest}/o) {
            $con->{"sock"} = $Accept;
            &send_client($con, &do_register($con));
            next CON;
        }
        if ($con->{"URL"} =~ /$main::Const{LoginRequest}/o) {
            $con->{"sock"} = $Accept;
            &send_client($con, &do_login);
            next CON;
        }

        #
        #  Normal request case.
        #
        if ($gb{neverfork}) {
            &debug('http', "Handling client request");
            # since we're ignoring SIGPIPE when in neverfork mode, protect the
            #  request processing from 'broken pipe' and similar system errors.
            &handle_conn($con);
        } elsif ($c = fork()) {
            # this is the parent
            $Children{$c}++;
            &debug('fork', "$Con: [$c]\n");
            $Accept->closesock;
        } elsif (! defined ($c)) {
            # fork failed!
            die "*** $$: Fatal: fork failed!\n";
        } else {
            # this is the child
            $SIG{PIPE} = \&child_cleanup;
            &handle_conn($con); 
            exit 0;
        }
    }
}


#--------------------------------------------------------------------
#  init other global state, and set up the @Config array to contain the
#  noncomment contents of the configuration file.
#

sub init {
    my @uname;

    @uname = &uname
        or die "Uname failed; $!";

    $SIG{'INT'} = $SIG{'KILL'} = 'parent_cleanup';
    #
    #  Read config info from file.
    #
    if (open(CFG, $g{config})) {
        @Config = ();
        while(<CFG>) {
            push(@Config, $_) unless /^#/;
        }
    } else {
        &log("Warning: can't read config file '$g{config}', using defaults");
    }
    #
    #  For all config entries except 'prefs', set them up as a hash.
    #
    my %cfg;
    for (@Config) {
        $cfg{$1} = $2 if ( ! /^prefs/
                          && /^\s*(\S+)\s+(.*)$/ );
    }
    #
    #  Initialize prefs manager.
    #
    &prefs::init;
    #
    #  Read clients.
    #
    &client::read_clients;
    #
    #  Initialize the refinement-icon imagemap stuff.
    #
    &ismap_icon::init;
    #
    # Make a temporary directory for segment files.
    #
    $main::Tempdir = $cfg{"tempdir"} || "/tmp";

    $main::Tempdir = "/tmp"
        unless (-d $main::Tempdir) && (-w $main::Tempdir);
    $main::Tempdir .= "/pyth$$";
    mkdir($main::Tempdir, 0700) or die "Can't make $main::Tempdir: $!";
    &debug('seg', "Made directory $main::Tempdir for page segment files");
    #
    #  Make a temporary directory for cache.  (For now, we use the same dir for
    #  both seg files and cache files.)
    #
    $main::Cache = new cache ($main::Tempdir,
                              (sub { substr(MD5->hexhash(shift),0,8) }),
                              10e6);
    #
    #  Where to find distillation programs.
    #
    $main::Bin = $cfg{"bin"} || "/bin";
    #
    #  Read the parser actions file.  If none was given in the config,
    #  use a default.
    #
    my $actions = $cfg{"html_actions"} || "./munge_html.pl";
    eval { require $actions };
    if ($@) {
        &main::log("Warning: can't read $actions ($@); using default");
        require "munge_html.pl";
    }

    my $host = ($Listen->endpoints)[2];
    my $rev = '$Id: pythia.pl,v 1.3 1997/02/26 09:31:41 fox Exp $';  #'
    $rev = $1 if $rev =~ /,v\s+([\d.]+)/;
    my $uptime = `/bin/date`;
    chomp $uptime;
    %main::Const = (%main::Const,
                    (
                    Crlf => "\r\012",
                    HTTPVersion => "HTTP/1.0",
                    Uptime => $uptime,
                    Rev => $rev,
                    Arch => $uname[0],
                    Hostname => $uname[1],
                    Proxy => 'Pythia',
                    SelectTimeout  => undef,
                    Ppid => $$,
                    RegisterRequest => "http://PythiaPrefs",
                    LoginRequest =>    "http://PythiaLogin",
                    StatsURL =>        "http://proxystats/?",
                    ProxyHost => $host,
                    ProxyPort => ($Listen->endpoints)[3],
                    GenerateRegisterURL =>
                    "http://$host:$g{port}/PythiaPrefs/%s/",
                    MatchRegisterURL =>
                    "http://$host:$g{port}/PythiaPrefs/([^/]+)/",
                    LoginURL =>
                    "http://$host:$g{port}/PythiaLogin/",
                    ));
    #
    #  initialize registration/login stuff.
    #
    &register::init;
    #
    #  Start services collector.
    #
    &distiller::startdefault($main::Const{Hostname},
                             $main::Const{ProxyPort},
                             $main::Const{Arch});
    1;
}

#----------------------------------------------------------------------
#  cleanup and die
#

sub parent_cleanup {
    &main::log("*** Proxy shutting down\n");
    &checkpoint;                            # sync prefs file before leaving
    unless (exists $main::Debug{'keeptemp'}) {
        &main::log( "*** Cleaning temp directory $main::Tempdir\n");
        system("/bin/rm -rf $main::Tempdir");
    }
    eval {$Listen->killsock()};
    eval {$Accept->killsock()};
    eval {$Cache->trash()};                 # delete page cache 
    eval {&distiller::killall()};            # shut down distillation svrs
    exit 0;
}

#---------------------------------------------------------------------------
# parse cmd line
#

sub parse_args {
    local(@debug);
    &usage if (grep(/-help/, @ARGV));

    # setup default args first

    while ($_ = shift @ARGV) {

        s/^-//;
        $g{$_} = shift @ARGV, next if defined $g{$_};
        $gb{$_} = 1, next if defined $gb{$_};
        $gb{$_} = 0, next if (s/^no// && defined $gb{$_});
        if (/^debug/) {
            @debug = split(/\s*,\s*/, shift @ARGV);
            if (grep (/^all$/, @debug)) {
                &main::log("All debugging messages enabled");
                $main::DebugAll = 1;
            } else {
                for (@debug) { $Debug{$_}++; }
                local($") = ",";                #" 
                &log("Debugging enabled for @debug");
            }
            next;
        }
        &main::log( "$0: arg '$_' ignored\n");
    }
}

#------------------------------------------------------------
#  Log
#

sub debug {
    local($key, @args) = @_;
    return unless ((exists $main::Debug{$key}) || $main::DebugAll);
    print STDERR "$$" . "[$key]: ";
    local($") = "\n";                       # "
    print STDERR "@args";
    print STDERR "\n" unless ((pop @args) =~ /\n$/);
}

sub log {
    local(@args) = @_;
    local($") = "\n";                       #" 
    print STDERR "$$:@args";
    print STDERR "\n" unless ((pop @args) =~ /\n$/);
}

sub get_stats {
    # Return a string (possibly including newlines) of interesting stats
    my $con = shift;
    my %prefs = &get_prefs($con->{"IPaddr"});
    my @x;

    push(@x, "Pythia (pid=$$) served $Con requests since $main::Const{Uptime}");
    my($bw,$lat) = &advice::estimate_bandwidth_latency(\%prefs,
                                                       $con->{"client"});
    push(@x, sprintf("Current bandwidth = %d, latency = %d", $bw, $lat));
    push(@x, sprintf("%d registered users (%d new this session)",
                     &prefs::register_info));
    my $svcs = &distiller::listall();
    push(@x, "Services:");
    push(@x, $svcs);
    push(@x, '');
    return join("<p>\n", @x);
}

#------------------------------------------------------------------
#  usage
#

sub usage {
    print STDERR "$0: HTTP proxy\n\n";
    print STDERR "  Usage: $0 [options]\n\n";

    print STDERR
        "-debug key[,key,...]: print debugging messages related to 'key':\n";
    print STDERR "                  advice - event notifications received\n";
    print STDERR "                  imgmap - server-side image map munging\n";
    print STDERR "                  stats - bytes saved by distillation\n";
    print STDERR "                  reg - login/prefs changes\n";
    print STDERR "                  refine - distillation/refinement requests\n";
    print STDERR "                  seg - page segmentation\n";
    print STDERR "                  conn - HTTP control traffic\n";
    print STDERR "                  echo - HTTP data traffic\n";
    print STDERR "                  echoall - all content traffic (verbose!)\n";
    print STDERR "                  fork - child process spawning/reaping\n";
    print STDERR "                  all - everything (really verbose!)\n";
    foreach (keys %g) {
        print STDERR "-$_: $help{$_} (default $g{$_})\n";
    }
    if (%gb) {
        print STDERR "\n  Boolean options (specify -foo or -nofoo):\n";
        foreach (keys %gb) {
            printf STDERR ("-$_: $help{$_} (def -%s$_)\n",
                           $gb{$_}? "": "no");
        }
    }
    exit 0;
}
